// Copyright (C) 2006-2010 Xilinx, Inc.  All rights reserved
// $Id: ArcCost.h,v 1.9 2010/11/08 19:03:09 jcorbett Exp $

// R#check_irs_separation - This file defines the interface to the
// Virtex-4 fault cost function with information from the Virtex-4
// Failure Analysis Report RPT059.

#ifndef Iv_ArcCost_h
#define Iv_ArcCost_h

#include <vector>
#include <map>
#include <utility>

#include "TUtilities/export/regex.hpp"

#include "RevArc.h"

// Pure virtual class to define the cost function interface, so that
// device-specific cost functions can be defined cleanly.
class ArcCost {
public:
   virtual int BitCount(RevArc& arc) = 0;
   virtual ~ArcCost();
};

class Virtex4Cost : public ArcCost {
public:
   Virtex4Cost();

   // The bit count is the cost of getting off the first node of the
   // arc and onto the second even signals do not flow that direction
   // in a correctly functioning device.
   int BitCount(RevArc& arc);
private:
   // The kind of physical structure associated with a node is not
   // publicly documented. It is found in
   // env/Databases/Rf/facade/virtex4/data/rf.acd.
   // Quoting from that file:
   //   "Names for the wire types defined by costcodes, the first entry is for
   //    costcode, typecode in rf talk, 0 the last is for costcode 24.
   //        This data copied from the NC_INFO.acd file"
   //::TYPECODENAMES=
   enum TypeCode {
      GENERIC      = 0,
      DOUBLE       = 1,
      INPUT        = 2,
      VUNIHEX      = 3,
      CNTRLPIN     = 4,
      CLKPIN       = 5,
      GLOBAL       = 6,
      OUTPUT       = 7,
      PINFEED      = 8,
      BOUNCEIN     = 9,
      KVCCOUT      = 10,
      IOBOUTPUT    = 11,
      OMUX         = 12,
      VLONG        = 13,
      OUTBOUND     = 14,
      HLONG        = 15,
      PINBOUNCE    = 16,
      BUFGOUT      = 17,
      BUFROUT      = 18,
      UNUSED       = 19,
      IOBIN2OUT    = 20,
      HUNIHEX      = 21,
      IOBINPUT     = 22,
      PADINPUT     = 23,
      PADOUTPUT    = 24,
      HVCCOUT      = 25,
      HGNDOUT      = 26
   };
   typedef boost::regex WireNamePat;
   typedef struct {
      const char* sourcePat;
      const char* targetPat;
      const Port::UINT32 cost;
   }
   CostPat;

   typedef std::vector<std::pair<WireNamePat, Port::UINT32> > WireCostList;
   typedef WireCostList::iterator WireCostListIt;
   typedef std::pair<WireNamePat, WireCostList> ArcCost;
   typedef std::vector<ArcCost> ArcCostList;
   typedef ArcCostList::iterator ArcCostListIt;
   // R#define_fence
   class ArcCostMatchSource {
   public:
      ArcCostMatchSource(const WireNamePat& target) : target(target) {}
      bool operator()(const ArcCost& a) const
         {
            return a.first == target;
         }
   private:
      const WireNamePat& target;
   };
private:
   // Convert regular expressions represented as character strings to
   // an internal representation for faster pattern matching.
   void CompileCostPatterns();

   int MatchBusWire(RevArc& arc,
                    const char* prefix,
                    Port::UINT32 prefixLen,
                    long busMax,
                    const char* suffix = "",
                    Port::UINT32 suffixLen = 0U);

public:
   std::vector<bool> sourceFilter0;
   std::vector<bool> sourceFilter1;
   std::vector<bool> targetFilter0;
   std::vector<bool> targetFilter1;
   ArcCostList arcCostList;
};

#endif
