#!/tools/xgs/perl/5.8.5/bin/perl -w
# 
# Copyright (C) 2007 Xilinx, Inc. All rights reserved.
# $Id: incver.pl,v 1.16.206.1 2013/07/17 18:44:01 jcorbett Exp $
#
# Generate the raw data for the Software Version Description document
# and the application version information by extracting RCS Id
# keywords from the source files. Output a file suitable for direct
# inclusion in the application source.
#
# This file defines three static strings for the application: version,
# iseBuild, and iseRelease.
#
# If any of the RCS $Id lines have changed since version.inc was
# written previously, update the file version.inc by incrementing the
# minor version number and refreshing the RCS $Id lines. If any files
# are checked out for editing, append 'a' to the version string. 
#
# To update the major version, simply edit the major version in the
# generated file.
#
# To force the version file to be updated, modify any of the RCS Id
# lines and optionally the version numbers.
#
# Usage: incver.pl <ise_build> <ise_product> <ise_version>
# For example,
#    incver.pl L.59 ISE 11.3
use strict;
# Flush output immediately so stdout and stderr are written in order.
$|++;

# Collect $Id lines among the sources and headers. Note if any file is
# writeable by assigning 'a' to $alphaStatus.

# There is something tricky things about this code. The RCS
# information we store in version.inc must not get changed by RCS when
# we check in version.inc. The actual RCS $Id keywords for this script
# and for versions.inc ought to still work. And we need to be able to
# distinguish our stored information from ordinary comments. So, our
# stored information is tagged with 'Id:' without the dollar sign and
# without 'Exp $' at the end of the line.

die 'Expecting <ise_build> <ise_product> <ise_version>' if !defined $ARGV[2];

my $overrideVersion = '';
if (defined $ARGV[3]) {
    $overrideVersion = $ARGV[3];
}

my $versionFile = 'version.inc';
my $alphaStatus = '';
my $ids = '';
my $idPat = '^// \$' . 'Id: (.*) Exp \$$'; #'# Concatenate to avoid RCS expansion.
my $oldIdPat = '^// Id: (.*)';
my $versionPat = '^static const char version\[\] *= *\"(\d+)\.(\d+)(.*)\";';
my $idCount = 0;
foreach my $f (<[is]/*.{h,inc,c}>) {
    open(FH, "<", "$f");
    binmode(FH);
    while (<FH>) {
        if ($_ =~ m/$idPat/) {
            # Collect RCS $Id line without the bit that triggers RCS
            # keyword expansion.
            $ids .= "// Id: $1\x0A";
            ++$idCount;
            last;
        }
    }
    close(FH);
    if ($f =~ m/.*\.[hc]/ && -w $f) {
        $alphaStatus = 'a';
        print "$f is writable.\n";
    }
}
die('Unable to find any RCS Ids') unless (0 < $idCount);

# Sort here because the relative paths above include directories.
$ids = join("\n", sort {uc($a) cmp uc($b)} split(/\n/, $ids)) . "\n";

# Collect RCS $Id lines from old versions file or if it does not
# exist, generate it.

my ($major, $minor, $oldAlpha) = (1, 0, '');
my $gotOldVersion = 0;
if (-e $versionFile) {
    my $oldIds = '';
    open(VER, "<", "$versionFile");
    binmode(VER);
    while (<VER>) {
	if ($_ =~ m/$oldIdPat/) {
	    # Collect RCS $Id line without the bit that triggers RCS
	    # keyword expansion.
	    $oldIds .= "// Id: $1\x0A";
	} elsif ($_ =~ /$versionPat/) {
	    # Note old version.
	    ($major, $minor, $oldAlpha) = ($1, $2, $3);
	    $gotOldVersion = 1;
	}
    }
    close(VER);
    die('Unable to find previous version') unless ($gotOldVersion);
    if ($overrideVersion =~ /(\d+)\.(\d+)(.*)/) {        
        ($major, $minor) = ($1, $2);
        &writeVersionInc($major, $minor, '');
    } elsif ($alphaStatus eq '' && $oldAlpha eq 'a') {
	# Everything is checked in.
	&writeVersionInc($major, $minor, '');
    } elsif ($oldIds ne $ids || ($alphaStatus eq 'a' && $oldAlpha eq '')) {
	# The ids changed or we just started a new alpha.
	&writeVersionInc($major, $minor + 1, $alphaStatus);
    } else {
	# The ids have not changed since last time and
	# the alpha status has not changed since last time.
	# But maybe the ISE version did change.
	&writeVersionInc($major, $minor, $alphaStatus);
    }
} else {
    # First time.
    &writeVersionInc(1, 0, $alphaStatus);
}



sub writeVersionInc {
    my ($major, $minor, $alphaStatus) = @_;
    if (substr($minor, 0, 1) ne '0' && $minor < 10) {
	$minor = '0' . $minor;
    }
    my @localTime = localtime(time());
    my $year = $localTime[5] + 1900;
    open(VER, ">", "$versionFile")
	|| die("$0: Unable to write $versionFile.\x0A" .
	       "Perhaps it needs to be unlocked for editing.");
    binmode(VER);		# Unix line endings on all platforms!
    print VER
	"// \$" . "Id: version.inc,v 1.0 2007/1/1 00:00:00 " .
	    "jcorbett Exp \$\x0A" .
	"// Copyright (C) $year Xilinx, Inc. All rights reserved.\x0A" .
	"// This file was automatically generated by $0.\x0A" .
	"static const char version[] = \"$major.$minor$alphaStatus\";\x0A" .
	"static const char iseBuild[] = \"$ARGV[0]\";\x0A" .
	"static const char iseRelease[] = \"$ARGV[1] $ARGV[2]\";\x0A\x0A" .
        $ids;
    close(VER);
}
