// Copyright (C) 2012-2013 Xilinx, Inc.  All rights reserved
// $Id: DiagramEmitter.c,v 1.5.2.20 2013/07/11 01:02:35 jcorbett Exp $
#include "DiagramEmitter.h"
#include "ColorTable.h"
#include "Svg.h"
#include "TUtilities/export/regex.hpp"
#include <algorithm>

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.5.2.20 $";

// Handle Spartan-6 irregularly and variously shaped GTPs.
// These are the only non-rectangular supersites in any of the
// supported devices.
bool IsRectangular(const Rf_Cell site) {
   const string siteType(site.GetCellDef().GetName());
   return siteType != "GTPA1_DUAL";
}

const double DiagramEmitter::areaRangeOutlineWidth2(0.05);
const double DiagramEmitter::areaRangeCornerRadius(0.2);
const double DiagramEmitter::areaRangeStrokeOpacity(0.5);
const double DiagramEmitter::areaRangeStrokeWidth(0.2);
const double DiagramEmitter::boxHeight(1.0);
const double DiagramEmitter::boxWidth(1.0);
const double DiagramEmitter::busMacroStrokeOpacity(0.73);
const double DiagramEmitter::tileStrokeOpacity(0.7);
const double DiagramEmitter::unusedTileStrokeOpacity(0.1);
const double DiagramEmitter::fillOpacity(0.4);
const double DiagramEmitter::fontSize(0.25);
const double DiagramEmitter::globalInset(0.15);
const double DiagramEmitter::insetOpacity(0.2);
const double DiagramEmitter::insetStrokeWidth(0.05);
const double DiagramEmitter::lineHeightSmall(0.23);
const double DiagramEmitter::lineHeightLarge(0.79);
const double DiagramEmitter::progOpacity(0.6);
const double DiagramEmitter::strokeWidth(0.1);
const double DiagramEmitter::specialInset(globalInset + strokeWidth);
const double DiagramEmitter::specialLeftMargin(0.5 * boxWidth);
const double DiagramEmitter::superSiteStrokeOpacity(0.2);
const double DiagramEmitter::textBaselineOffset(0.2);
const double DiagramEmitter::textLeftMargin(0.05);
const double DiagramEmitter::unusedStrokeWidth(0.05);
const string DiagramEmitter::globalStyle(
   Svg().Decl("fill", "none") +
   Svg().Decl("stroke", "black") +
   Svg().Decl("stroke-opacity", insetOpacity) +
   Svg().Decl("stroke-width", insetStrokeWidth));
const string DiagramEmitter::font(Svg().Attr("font-size", fontSize) +
                                  Svg().Attr("font-family", "Helvetica"));
const string DiagramEmitter::logicStrokeColor("black");
const string DiagramEmitter::unusedStyle(
   Svg().Decl("fill", "none") +
   Svg().Decl("stroke", logicStrokeColor) +
   Svg().Decl("stroke-opacity", 0.5 * superSiteStrokeOpacity) + 
   Svg().Decl("stroke-width", insetStrokeWidth));
const string DiagramEmitter::superSiteStyle(
   Svg().Decl("fill", "url(#stipple)") +
   Svg().Decl("stroke", logicStrokeColor) +
   Svg().Decl("stroke-opacity", superSiteStrokeOpacity) + 
   Svg().Decl("stroke-width", insetStrokeWidth));
const string DiagramEmitter::superSiteViolation(
   Svg().Decl("fill", "url(#crosshatch)") +
   Svg().Decl("fill-opacity", 1.0) +
   Svg().Decl("stroke", "none"));
const string DiagramEmitter::shortStyle(
   Svg().Decl("fill", "url(#horstripes)") +
   Svg().Decl("stroke", logicStrokeColor) +
   Svg().Decl("stroke-opacity", superSiteStrokeOpacity) + 
   Svg().Decl("stroke-width", insetStrokeWidth));
const string DiagramEmitter::vertStyle(
   Svg().Decl("fill", "url(#vertstripes)") +
   Svg().Decl("stroke", logicStrokeColor) +
   Svg().Decl("stroke-opacity", superSiteStrokeOpacity) + 
   Svg().Decl("stroke-width", insetStrokeWidth));

// Since the config center area is not associated with a single unique
// site, information needed to render this area is handled specially.
void DiagramEmitter::SetHasConfigCenterViolation() {
   hasConfigCenterViolation = true;
}

// Unfortunately, there are some sites that this code is unable to
// render in a useful way. Since the interconnect blocks associated
// with a site are essential in determining the extent of the site, if
// a site is not directly associated with any interconnects or the
// interconnects are spread across many rows or columns, drawing the
// site won't work out. All the sites needed for analysis are either
// well-behaved or handled as special cases.
bool DiagramEmitter::IsDrawableSuperSite(const Rf_Cell superSite) {
   // No supersite rendering before Spartan-6.
   if (deviceModel->IsVirtex4() || deviceModel->IsVirtex5()) {
      return false;
   }
   ASSERT(superSite);
   const string siteType(superSite.GetCellDef().GetName());

   return !(deviceModel->IsSpartan6() &&
            (siteType == "IOBM" || siteType == "IOBS")) &&
      deviceModel->IsUserTileType(siteType) &&
      deviceModel->SpansMultipleTiles(superSite) ||
      (deviceModel->Is7Series() &&
       deviceModel->IsConfigCenterType(siteType));
}

const size_t smallZynqHeight(200);
const size_t zynqPssWidth(18);

DiagramEmitter::DiagramEmitter(Data& data,
                               const boost::shared_ptr<ContourMap<Tile> >&
                                  contourMap,
                               const TileVec& grid,
                               GroupSetFamily& groupSetFamily,
                               const DeviceModel::CellSet& superIntSet,
                               const BoundaryMap& superBoundaryMap,
                               const BoundaryMap& superGlobalMap,
                               Data::GroupSet*& configCenterGroupSet)
   : data(data),
     deviceModel(data.GetDeviceModel()),
     contourMap(contourMap),
     grid(grid),
     groupSetFamily(groupSetFamily),
     superIntSet(superIntSet),
     superBoundaryMap(superBoundaryMap),
     superGlobalMap(superGlobalMap),
     configCenterGroupSet(configCenterGroupSet),
     hasConfigCenterViolation(false),
     smallZynqOffset(deviceModel->IsZynq() &&
                     deviceModel->GetTileRowCount() == smallZynqHeight
                     ? zynqPssWidth : 0.0)
{}

// R#check_irs_separation, R#draw_fence - Color code the group
// lists. Singleton sets are assigned colors in alphabetical
// order. For example, if there were three isolation groups, A, B and
// C, they would be assigned gray, red, and blue, respectively.
//
// If the set is a singleton, use the next color from the list,
// unless there is no next color, then use the "overflow" color.
//
// If multiple isolated signals occupy a tile, color the tile bright 
// orange.
//
// There is no means to distinguish intended connections between
// isolated regions from accidental ones. All must be carefully
// examined.
//
// See ColorTable.h for color code and constants.
void DiagramEmitter::LabelGroupSets() {
   groupSetFamily.Sort();
   size_t z(groupSetFamily.groupSetList.size());
   size_t color(ColorName_User);
   colorMap.resize(z);
   for (size_t i(0); i < z; ++i) {
      size_t cardinality(groupSetFamily.groupSetList[i]->size());
      if (cardinality == 1) {
         colorMap[i] = color;
         // The last color in the list is the overflow color.
         // Only increment if there are more colors.
         if (color < ColorName_Size - 1) {
            ++color;
         }
      } else {
         // Multiple isolated signals in a tile.
         colorMap[i] = ColorName_Mixed;
      }
   }
}

// Return the CSS color associated with the given group set.  Use the
// position of this tile's group set in the master list of group sets
// to determine the tile's color.
const string DiagramEmitter::GetGroupColor(Data::GroupSet* gs) {
   ASSERT(!groupSetFamily.groupSetList.empty());
   ASSERT(gs && gs->size());
   size_t colorIndex(ColorName_Mixed);
   GroupSetFamily::GroupSetListIt beginIt(groupSetFamily.groupSetList.begin());
   GroupSetFamily::GroupSetListIt endIt(groupSetFamily.groupSetList.end());
   GroupSetFamily::GroupSetListIt gsIt(find(beginIt, endIt, gs));
   ASSERT(gsIt != endIt);
   if (gs->size() == 1) {
      size_t groupRank(gsIt - beginIt);
      ASSERT(groupRank < colorMap.size());
      colorIndex = colorMap[groupRank];
   }
   return string(colorTable[colorIndex].colorCode);
}

// Fill colors are nearly transparent unless the corresponding site
// belongs to multiple groups in which case the color is nearly
// opaque.
const string DiagramEmitter::GetGroupColorOpacity(Data::GroupSet* gs) {
   ASSERT(!groupSetFamily.groupSetList.empty());
   ASSERT(gs && gs->size());
   size_t colorIndex(ColorName_Mixed);
   GroupSetFamily::GroupSetListIt beginIt(groupSetFamily.groupSetList.begin());
   GroupSetFamily::GroupSetListIt endIt(groupSetFamily.groupSetList.end());
   GroupSetFamily::GroupSetListIt gsIt(find(beginIt, endIt, gs));
   ASSERT(gsIt != endIt);
   if (gs->size() == 1) {
      size_t groupRank(gsIt - beginIt);
      ASSERT(groupRank < colorMap.size());
      colorIndex = colorMap[groupRank];
   }
   return string(colorIndex == ColorName_Mixed ? "0.8" : "0.4");
}

// Get the tile pointer for a given cell.
// Note: It is an error to invoke this function on a cell that does
// not have a corresponding tile.
Tile* DiagramEmitter::GetTile(Rf_Cell cell) {
   Port::UINT32 column(cell.GetTileColumn());
   Port::UINT32 row(cell.GetTileRow());
   Tile* tile(contourMap->At(column, row));
   ASSERT(tile);
   return tile;
}

// Pattern to match coordinate labels where both x and y are multiples
// of 10. This is used to decimate the set of labels drawn, making the
// diagrams somewhat less cluttered.
const boost::regex decadePat(".*0,.*0");

// When we need this we're in the uncomfortable position of processing
// a part that contains one supersite the entire height of the
// die that only has interconnect on one edge. This defeats our
// approach of using the interconnects of a supersite to determine its
// visible bounds and our strategy to hide columns that are irrelevant
// for analysis.
double DiagramEmitter::GetSmallZynqOffset() {
   return smallZynqOffset;
}


// R#draw_fence - For each interconnect tile that contains at least
// one signal,
//    - Draw a square color-coded by isolation group.
//    - Label it with the interconnect tile coordinates.
//    - Mark tiles containing global signals with an inset square.
//    - Mark tiles containing routing or programming that violates
//      isolation rules with an X.
//    - Mark tiles containing bus macros with a black outline.
//    - Mark tiles with terminals with a small square in the middle,
//      but only for tiles without isolated logic or routing.
//    - Mark tiles containing slices with a vertical rectangle in the
//      right half of the tile.
//    - Mark tiles containing other logic with a slightly thinner
//      vertical rectangle, possibly spanning several rows for block
//      RAM, DSPs and other logic blocks that span multiple tiles.
//      Fill the non-slice logic with a stipple pattern to indicate
//      that it is used. Draw an empty outline otherwise.
//    - Mark invalid fence tiles that are adjacent directly or
//      indirectly to isolated resources with a diamond tinted by the
//      color of the isolated resource. When multiple isolation taint
//      a tile, the diamond is bright orange and the tile is marked
//      with an X.
const string DiagramEmitter::EmitTiles() {
   Svg d;
   string labels, boxes;
   Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   for (size_t i(0); i < grid.size(); ++i) {
      string fillColor(colorTable[ColorName_Blank].colorCode);
      string strokeColor("black");
      double strokeOpacity(unusedTileStrokeOpacity);
      const double x(contourMap->GetAdjustedX(i)
                     + GetSmallZynqOffset());
      const double y(contourMap->GetAdjustedY(i));

      Tile* tile(grid.at(i));
      if (tile) {
         Rf_Cell cell(tile->GetCell());

         // The PCIE has two columns of interconnects. To avoid
         // confusing end users, IVT does not show PCIE
         // interconnects.
         if (deviceModel->IsInterconnectCell(cell) &&
             !tile->HasSiteType("PCIE_2_1") &&
             !tile->HasSiteType("PCIE") &&
             !tile->HasSiteType("PCIE_A1"))
         {
            bool needsRect(true);
            // Mark PIPs with a square dot in the center of the tile.
            if (tile->HasTerminal()) {
               // It's not necessarily a violation, but we may need
               // more information when troubleshooting.
               const string pipStyle(d.Decl("fill", "black") +
                                     d.Decl("fill-opacity", 0.5));
               boxes +=
                  d.Rect(x + (boxWidth - globalInset)/2,
                         y + (boxHeight + globalInset)/2,
                         x + (boxWidth + globalInset)/2,
                         y + (boxHeight - globalInset)/2,
                         pipStyle)
                  + nl;
            }
            // We know the cell is good because we checked before we
            // inserted it in the grid.

            // Outline the Trusted Bus Macro tiles in black.
            if (tile->HasBusMacro()) {
               boxes += d.Comment("BusMacro: " + tile->GetName() + " " + tile->Describe());
               strokeOpacity = busMacroStrokeOpacity;
            }

            Data::GroupSet* gs(tile->GetGroupSet());
            bool needsLabel(false);
            bool ignoreOffset(false);
            if (gs) {
               fillColor = GetGroupColor(gs);
               strokeColor = fillColor;
               const string fillOpacity(GetGroupColorOpacity(gs));
               if (!data.NeedConstraintsCheck() || tile->HasBadContent()) {
                  strokeOpacity = tileStrokeOpacity;

                  // In UCF mode, we assume all tiles have nets
                  // but don't need every possible label. Whereas in
                  // NCD mode, we label all the tiles that have actual
                  // nets in them.
                  needsLabel = true;

                  if (tile->HasIsolatedNet() || tile->HasProgramming()) {
                     // In NCD mode, the tile color is based on
                     // logical hierarchy, not on geometric location
                     // with respect to an area range constraint.

                     // Draw an outlined square color-coded by isolation group.
                     boxes +=
                        d.Rect(x, y + boxHeight, x + boxWidth, y,
                               d.Decl("fill", fillColor) +
                               d.Decl("fill-opacity", fillOpacity) +
                               d.Decl("stroke", strokeColor) +
                               d.Decl("stroke-opacity", strokeOpacity) +
                               d.Decl("stroke-width", strokeWidth)) +  nl;
                     needsRect = false;
                  } else {
                     ;
                     // In UCF-mode netDescriptions will be empty. No need to
                     // draw the rectangle or add the label.
                  }
               }
            }
            if (tile->HasGlobalNet()) {
               needsLabel = true; // Has a global net.

               if (!tile->HasIsolatedNet() && !tile->HasProgramming()) {
                  // Draw an outline to indicate something is here.
                  boxes += d.Rect(x, y + boxHeight, x + boxWidth, y,
                                  d.Decl("fill", "none") +
                                  d.Decl("stroke", "black") +
                                  d.Decl("stroke-opacity", strokeOpacity) +
                                  d.Decl("stroke-width", strokeWidth))
                     + nl;
                  needsRect = false;
               }
               // Mark globals with an inset square.
               boxes +=
                  d.Rect(x + globalInset,
                         y + boxHeight - globalInset,
                         x + boxWidth - globalInset,
                         y + globalInset,
                         globalStyle)
                  + nl;
            }            
            if (!data.NeedConstraintsCheck() && tile->HasProgramming()) {

               // Draw tall a skinny rectangle in the right half of
               // the column to indicate the presence of logic. 
               Rf_Cell logicSite;
               Rf_Cell prevTop, prevBottom;
               for (DeviceModel::CellSetIt csIt(tile->logicSiteSet.begin());
                    csIt != tile->logicSiteSet.end();
                    ++csIt)
               {
                  logicSite = *csIt;
                  ASSERT(logicSite);

                  Rf_Cell intTileTop(deviceModel->GetIntTile(logicSite,
                                                             tileDef,
                                                             -1));
                  Rf_Cell intTileBottom(deviceModel->GetIntTile(logicSite,
                                                                tileDef,
                                                                1));
                  if (prevTop == intTileTop &&
                      prevBottom == intTileBottom)
                  {
                     break; // Same extent, no need to redraw.
                  }

                  prevTop = intTileTop;
                  prevBottom = intTileBottom;

                  size_t iy0(intTileBottom.GetTileRow());
                  size_t iy1(intTileTop.GetTileRow());
                  size_t siteRows(iy1 - iy0 + 1);
                  string siteType(logicSite.GetCellDef().GetName());
                  if (intTileBottom == cell &&
                      IsPrefix("SLICE", siteType.c_str()))
                  {
                     // The logic here supports drawing of tall sites,
                     // but currently only slices are drawn this way.

                     // Supersites are drawn elsewhere.
                     boxes += d.Comment("Logic: " + tile->Describe());
                     boxes +=
                        d.Rect(x + specialLeftMargin,
                               y + siteRows * boxHeight - specialInset,
                               x + boxWidth - specialInset,
                               y + specialInset,
                               d.Decl("fill", fillColor) +
                               d.Decl("fill-opacity", progOpacity) +
                               d.Decl("stroke", strokeColor) +
                               d.Decl("stroke-width", insetStrokeWidth))
                        + nl;
                     needsLabel = true;
                     ignoreOffset = true;
                     needsRect = false;
                  }
               }
            }
            if (tile->HasViolation()) {
               // If a tile contains a wire that has a terminal
               // (used or unused) in the fence, there is an
               // isolation failure. Note: HasTerminal() is only
               // set for interconnect tiles in the fence when the
               // net is not driven by a global clock.
               //
               // If a tile containing isolated nets from one group
               // is adjacent to a tile containing isolated nets
               // from another group, there is an isolation
               // failure.
               //
               // If a single tile contains isolated nets from more
               // than one group, there is an isolation failure.
               //
               // Draw an X through the box.
               boxes += d.Comment("Violation: " + tile->Describe()) + nl;
               boxes += d.XRect(x, y + boxHeight, x + boxWidth, y) + nl;
               needsLabel = true; // Has a violation.
            }
            if (needsRect && !superIntSet.count(tile->GetCell())) {
               // Outline tiles that are unused.
               boxes += d.Rect(x, y + boxHeight, x + boxWidth, y,
                               d.Decl("fill", "none") +
                               d.Decl("stroke", "black") +
                               d.Decl("stroke-opacity", unusedTileStrokeOpacity) +
                               d.Decl("stroke-width", unusedStrokeWidth))
                  + nl;
            }

            string xyLabel(tile->GetLabel());
            const bool isDecade(boost::regex_match(xyLabel, decadePat));
            if (needsLabel || isDecade) {
               // Label the cell. Stagger the vertical position of the
               // labels based on the column mod 4. This reduces the
               // number of labels that land on top of one another.

               // Stagger the y-offset of the labels within the tile.
               const int staggerCount(4);
               double modOffset((int(x) % staggerCount) * lineHeightSmall);
               if (isDecade) {
                  modOffset = 0.0;
               }
               if (ignoreOffset) {
                  // cell names
                  if (!tile->logicSiteSet.empty()) {
                     string sep;
                     for (DeviceModel::CellSetIt
                             csIt(tile->logicSiteSet.begin());
                          csIt != tile->logicSiteSet.end();
                          ++csIt)
                     {
                        string siteName(csIt->GetName());
                        if (!IsPrefix("SLICE", siteName.c_str())) {
                           xyLabel += sep + string(siteName);
                           sep = ", ";
                        }
                     }
                  }
                  modOffset = lineHeightSmall;
               }
               labels += d.Text(x + textLeftMargin,
                                y + textBaselineOffset + modOffset,
                                xyLabel) + nl;
            }
            if (tile->HasTaints()) {
               Data::GroupSet* tgs(tile->GetTaintGroupSet());
               string taintFill("black");
               string taintOpacity("0.106");
               if (tgs) taintFill = GetGroupColor(tgs);
               if (tgs && tile->HasMultipleTaints()) taintOpacity = "0.76";
               boxes += d.Comment("Taint mark: " + tile->Describe()) + nl;
               // Diamond
               CoordVec boundary;
               double h(boxHeight/4.0);
               double w(boxWidth/4.0);

               boundary.push_back(x + w);
               boundary.push_back(y + h + h);

               boundary.push_back(x + w + w);
               boundary.push_back(y + h);

               boundary.push_back(x + w + w + w);
               boundary.push_back(y + h + h);

               boundary.push_back(x + w + w);
               boundary.push_back(y + h + h + h);

               boxes +=
                  d.Polygon(boundary,
                            d.Decl("fill", taintFill) +
                            d.Decl("fill-opacity", taintOpacity))
                  + nl;               
            }
         }
      }
   }
   labels = d.Elt("g", font, labels);
   return d.Comment("Tiles") + nl +
      boxes + nl +
      labels + nl;
}

// Not currently used. This class helps visualize the relationship
// between logic sites and the interconnect blocks directly associated
// with them.
class DebugInts {
public:
   DebugInts(const boost::shared_ptr<ContourMap<Tile> >& contourMap,
             double x, double y,
             std::string& rays)
      : contourMap(contourMap), x(x), y(y), rays(rays) {};
   void operator()(Rf_Cell cell) {
      Svg d;
      string textStyle(d.Attr("fill-opacity", 1.0) + d.Attr("font-size", 0.1));
      size_t i(contourMap->Index(cell.GetTileColumn(),
                                 cell.GetTileRow()));
      // Note: does not include GetSmallZynqOffset().
      double tx(contourMap->GetAdjustedX(i));
      double ty(contourMap->GetAdjustedY(i));
      rays += d.Line(x, y, tx, ty,
                     d.Decl("stroke", "blue") +
                     d.Decl("stroke-width", 0.005) +
                     d.Decl("stroke-opacity", 0.1)) + '\n';
      string asterisk(tx < x ? "*" : "");
      rays +=
         d.Text(tx + 0.1, ty + 0.7,
                string(cell.GetName()) + asterisk,
                textStyle) + '\n';
   }
   const boost::shared_ptr<ContourMap<Tile> >& contourMap;
   double x;
   double y;
   std::string& rays;
};

// Add a label at a given location in a given style. Multiple labels
// at a given location are stored in a set and eventually listed
// separated by commas. In some cases, labels are translated to
// general terms, e.g. a "RAMB18" is rendered as "Block RAM".
void DiagramEmitter::AddLabel(double x, double y, string text, string style) {
   ASSERT(0 <= x);
   ASSERT(0 <= y);
   ASSERT(!text.empty());
   ASSERT(x < 10000);
   ASSERT(y < 10000);
   Label label = { x, y, text, style };
   size_t locationHash(10000 * static_cast<size_t>(10 * y) +
                       static_cast<size_t>(10 * x));
   LabelMapIt lsIt(labelMap.find(locationHash));
   if (lsIt == labelMap.end()) {
      LabelSet ls;
      ls.insert(label);
      labelMap.insert(make_pair(locationHash, ls));
   } else {
      lsIt->second.insert(label);
   }
}

// Emit labels combining some under more general term.
const string DiagramEmitter::EmitLabels() {
   Svg d;
   string result("");
   string sep("");
   for (LabelMapIt lmIt(labelMap.begin()); lmIt != labelMap.end(); ++lmIt) {
      LabelSet& ls(lmIt->second);
      ASSERT(!ls.empty());
      double x(ls.begin()->x);
      double y(ls.begin()->y);
      string style(ls.begin()->style);
      string labelList;
      sep = "";
      for (LabelSetIt lsIt(ls.begin()); lsIt != ls.end(); ++lsIt) {
         if (IsPrefix("RAM", lsIt->text.c_str())) {
            labelList = "Block RAM";
            break;
         } else if (IsPrefix("DSP", lsIt->text.c_str())) {
            labelList = "DSP";
            break;
         } else if (lsIt->text == "IN_FIFO") {
            labelList = "I/O Fifo";
            break;
         }
         labelList += sep + lsIt->text;
         sep = ", ";
      }
      result += d.Text(x, y, labelList, style) + nl;
   }
   return result;
}

// If site is a 7 series IOB, draw it. This is needed because there
// are two kinds of IOB in 7 series, singles and pairs.
void DiagramEmitter::DrawIOBSite(Rf_Cell site, string& boxes, string& labels) {
   Svg d;
   const Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
   string siteType(site.GetCellDef().GetName());
   if (deviceModel->Is7Series() &&
       deviceModel->IsUserTileType(siteType) &&
       IsPrefix("IOB", siteType.c_str()) &&
       !deviceModel->SpansMultipleTiles(site))
   {
      Rf_Cell cell(deviceModel->GetSiteParentTile(site));
      size_t i(contourMap->Index(cell.GetTileColumn(),
                                 cell.GetTileRow()));
      double tx(contourMap->GetAdjustedX(i) + GetSmallZynqOffset());
      double ty(contourMap->GetAdjustedY(i));

      double offset(4 * 0.15 + 0.1);

      double rLeft(tx + offset);
      double rTop(ty + boxHeight - specialInset);
      double rRight(tx + offset + 0.15);
      double rBottom(ty + specialInset);

      Tile* tile(GetTile(cell));
      Data::GroupSet* gs(tile->GetGroupSet());
      const bool isNcd(!data.NeedConstraintsCheck());
      string style(isNcd && tile->HasProgramming() ? superSiteStyle : unusedStyle);
      // In UCF-mode, sites are not colored unless they are in
      // multiple isolation groups. In NCD-mode, the actual site must
      // be configured and the color is determined by module
      // hierarchy.
      const bool needsFill(tile->HasBadContent() ||
                           (gs &&
                            isNcd &&
                            tile->HasProgramming()));
      const string fillColor(needsFill ? GetGroupColor(gs) : "none");
      const string fillOpacity(needsFill ? GetGroupColorOpacity(gs) : "0");

      boxes += d.Comment("Site[1]: " + siteType) + nl +
         d.Rect(rLeft, rTop, rRight, rBottom,
                d.Decl("fill", fillColor) +
                d.Decl("fill-opacity", fillOpacity) +
                d.Decl("stroke", "none")) + nl;
      boxes += d.Rect(rLeft, rTop, rRight, rBottom, style) + nl;
      if (tile->HasViolation()) {
         boxes +=
            d.Comment(tile->Describe()) + nl +
            d.Rect(rLeft, rTop, rRight, rBottom, superSiteViolation)
            + nl;
      }

   }
}

// The 7 series CMT column contains many tall sites that overlap one
// another. The patterns below are used to identify the sites so they
// can be drawn in specific subcolumns within the CMT column.

const boost::regex fifoPat(".*_FIFO");
const boost::regex inPat("PHASER_IN.*");
const boost::regex outPat("PHASER_OUT.*");
const boost::regex refPat("PHASER_REF.*");
const boost::regex mmcmPat("MMCM.*");
const boost::regex pllPat("PLL.*");

// Return the horizontal offset within the CMT column of the given
// siteType.
double DiagramEmitter::GetSiteOffset(const string siteType) const {
   double offset(0.1);

   // EFUSE_USR is unique per device. Use it to represent the config
   // center column.
   const bool isConfigCenter(deviceModel->Is7Series() &&
                             siteType == "EFUSE_USR");
   const bool isZynqPSS(deviceModel->Is7Series() &&
                        siteType == "PS7");
   if (boost::regex_match(siteType, fifoPat)) {
      offset = 0.1;
   } else if (boost::regex_match(siteType, inPat)) {
      offset = 1 * 0.15 + 0.1;
   } else if (boost::regex_match(siteType, outPat)) {
      offset = 2 * 0.15 + 0.1;
   } else if (boost::regex_match(siteType, refPat)) {
      offset = 3 * 0.15 + 0.1;
   } else if (boost::regex_match(siteType, mmcmPat)) {
      // The case is broken out in case we want to
      // change the way the MMCM is rendered.
      offset = 4 * 0.15 + 0.1;
   } else if (boost::regex_match(siteType, pllPat)) {
      offset = 4 * 0.15 + 0.1;
      // This case is broken out in case we want to
      // change the way the PLL is rendered.
   } else if (isConfigCenter) {
      offset = 0.1;
   } else if (isZynqPSS) {
      offset = 0.1;
   } else { // Other logic sites
      offset = 4 * 0.15 + 0.1;
   }

   return offset;
}

// Indicate the 7 series clock column with vertical hatching. This
// column cannot be used as a vertical fence.
void DiagramEmitter::DrawClockColumn(string& boxes) {
   Svg d;
   if (deviceModel->Is7Series()) {
      Rf_Cell site;
      string siteType;
      Rf_CellContainer sites(deviceModel->GetSites());

      // For each site
      for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
           sitesIt != sites.end();
           ++sitesIt)
      {
         site = *sitesIt;
         siteType = site.GetCellDef().GetName();
         if (boost::regex_match(siteType, deviceModel->clockBufferPat)) {
            break;
         }
      }

      const Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());

      // Get the INT cells at the corners of the site.
      BoundInts bb;
      deviceModel->MapIntTiles(bb, site, tileDef);

      // Get display coordinates from the corner cells.
      size_t iLeft(contourMap->Index(bb.left.GetTileColumn(),
                                     bb.left.GetTileRow()));
      double xLeft(contourMap->GetAdjustedX(iLeft)
                   + GetSmallZynqOffset());

      size_t iRight(contourMap->Index(bb.right.GetTileColumn(),
                                      bb.right.GetTileRow()));
      double xRight(contourMap->GetAdjustedX(iRight)
                    + GetSmallZynqOffset());

      const long height(contourMap->AdjustY(contourMap->RowCount() - 1) + 1);
      boxes += nl + d.Comment("Clock column") +
         nl + d.Rect(xRight, height * boxHeight,
                     xRight + boxWidth, 0,
                     vertStyle);
   }
}

// Supersites are sites that span multiple tiles, such as Block RAM,
// PCIE, and Gigabit Transceivers. Several properties of supersites
// are illustrated. Isolation group membership is indicated by fill
// color. Usage is indicated by a stipple pattern. Isolation violation
// is indicated by crosshatching. The presence of global routing in a
// supersite is indicated by an inset outline. The config center area
// is not modeled as a supersite, but this distinction is not helpful
// to users, so it is rendered as a supersite. The config center area
// cannot be used as a fence. This is suggested by the use of
// horizontal rules.
void DiagramEmitter::DrawRectangularSuperSite(Rf_Cell site,
                                              string& boxes, string& labels)
{
   Svg d;

   if (IsDrawableSuperSite(site) && IsRectangular(site)) {
      const Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());

      // Get the INT cells at the corners of the site.
      BoundInts bb;
      deviceModel->MapIntTiles(bb, site, tileDef);
      Rf_Cell cell(deviceModel->GetSiteParentTile(bb.top));
      Tile* tile(GetTile(cell));
      ASSERT(tile);

      // Get display coordinates from the corner cells.
      size_t iLeft(contourMap->Index(bb.left.GetTileColumn(),
                                     bb.left.GetTileRow()));
      double xLeft(contourMap->GetAdjustedX(iLeft)
                   + GetSmallZynqOffset());

      size_t iRight(contourMap->Index(bb.right.GetTileColumn(),
                                      bb.right.GetTileRow()));
      double xRight(contourMap->GetAdjustedX(iRight)
                    + GetSmallZynqOffset());

      size_t iTop(contourMap->Index(bb.top.GetTileColumn(),
                                    bb.top.GetTileRow()));
      double yTop(contourMap->GetAdjustedY(iTop));

      size_t iBottom(contourMap->Index(bb.bottom.GetTileColumn(),
                                       bb.bottom.GetTileRow()));
      double yBottom(contourMap->GetAdjustedY(iBottom));

      string siteType(site.GetCellDef().GetName());
      double offset(GetSiteOffset(siteType));

      // Tranform corner coordinates to supersite box edges.
      double rLeft(xLeft + offset);
      double rTop(yTop + boxHeight - specialInset);
      double rRight(xRight + offset + 0.15);
      double rBottom(yBottom + specialInset);

      Data::GroupSet* gs(tile->GetGroupSet());
      const bool isConfigCenter(deviceModel->Is7Series() &&
                                siteType == "EFUSE_USR");
      const bool isZynqPSS(deviceModel->Is7Series() &&
                           siteType == "PS7");
      if (isConfigCenter) {
         gs = configCenterGroupSet;
         rLeft -= 6.0 * boxWidth;
         rRight += boxWidth - 0.15 - specialInset;
         rTop += 46.0 * boxHeight;
         rBottom -= 50.0 * boxHeight;
      } else if (isZynqPSS) {
         rLeft -= zynqPssWidth * boxWidth + 0.1;
         rRight += boxWidth - offset - 0.15;
         rBottom -= 1.0 * boxHeight;
      }

      const bool isNcdMode(!data.NeedConstraintsCheck());
      const bool isMixed(gs && 1 < gs->size());
      // In UCF mode all sites within area ranges are assumed to be
      // used, so there's no point in stippling them.
      const bool usesSite(isMixed || (isNcdMode &&
                                      (tile->logicSiteSet.find(site) !=
                                       tile->logicSiteSet.end())));
      string style(usesSite ? superSiteStyle : unusedStyle);

      // Tile::HasBadContent() and Tile::HasBadNeighbor() are not
      // applicable because there is no single tile that is always
      // present when some config center site is used.
      const bool needsFill(usesSite && gs);
      const string strokeColor(gs ? GetGroupColor(gs) : "none");
      const string fillColor(needsFill ? strokeColor : "none");
      const string fillOpacity(needsFill ? GetGroupColorOpacity(gs) : "0");

      const string strokeStyle(
         d.Attr("style",
                d.Decl("fill", "none") +
                d.Decl("stroke-opacity", areaRangeStrokeOpacity) +
                d.Decl("stroke-width", areaRangeStrokeWidth)));
      if (isConfigCenter) {
         boxes += d.Elt("g", strokeStyle,
                        nl +
                        AreaRangeRect("Group: " + SetToString(gs, ",") +
                                      " Range: CONFIG CENTER (Mixed)",
                                      strokeColor,
                                      rLeft - offset, rTop + specialInset,
                                      rRight + offset, rBottom - specialInset));
         const long height(contourMap->AdjustY(contourMap->RowCount() - 1) + 1);
         if (smallZynqHeight < deviceModel->GetTileRowCount()) { 
            boxes += nl + d.Comment("Config center interconnect column") +
               nl + d.Rect(xRight, height * boxHeight,
                           xRight + boxWidth, 0,
                           vertStyle);
         }
         // Config Center logic box (an area, not one site)
         boxes += d.Comment("Site[4]: " + siteType) + nl +
            d.Rect(rLeft, rTop, rRight, rBottom, shortStyle) + nl;
      } else if (isZynqPSS) {

         bool needAreaRange(false);
         for (Data::SiteRectMapIt srmIt(data.siteRectMap.begin());
              srmIt != data.siteRectMap.end();
              ++srmIt)
         {
            DeviceModel::SiteRectVec rangeList(srmIt->second);
            for (DeviceModel::SiteRectVecIt rlIt(rangeList.begin());
                 rlIt != rangeList.end();
                 ++rlIt)
            {
               DeviceModel::SiteRect rc(*rlIt);
               if (rc.siteType == "PS7") {
                  needAreaRange = true;
                  break;
               }
            }
         }

         if (needAreaRange) {
            boxes += d.Elt("g", strokeStyle,
                           nl +
                           AreaRangeRect("Group: " + SetToString(gs, ",") +
                                         " Range: Zynq PSS",
                                         strokeColor,
                                         rLeft, rTop + specialInset,
                                         rRight, rBottom - specialInset));
            // Zynq PSS site
            boxes += d.Comment("Site[5]: " + siteType) + nl +
               d.Rect(rLeft, rTop, rRight, rBottom, shortStyle) + nl;
         } else {
            // If the area range is omitted, draw it anyway, but draw
            // it dashed orange and black.
            boxes += d.Elt("g", strokeStyle,
                           nl +
                           AreaRangeRect("Group: " + SetToString(gs, ",") +
                                         " Range: Missing Zynq PSS",
                                         "black",
                                         rLeft, rTop + specialInset,
                                         rRight, rBottom - specialInset));
            const string dashStyle(
               d.Attr("style",
                      d.Decl("fill", "none") +
                      d.Decl("stroke-opacity", "1.0") +
                      d.Decl("stroke-width", areaRangeStrokeWidth) +
                      d.Decl("stroke-dasharray", "0.5,0.5")));
            const string strokeColor(colorTable[ColorName_Mixed].colorCode);
            boxes += d.Elt("g", dashStyle,
                           nl +
                           AreaRangeRect("Group: " + SetToString(gs, ",") +
                                         " Range: Missing Zynq PSS",
                                         strokeColor,
                                         rLeft, rTop + specialInset,
                                         rRight, rBottom - specialInset));
            // Zynq PSS site
            boxes += d.Comment("Site[6]: " + siteType) + nl +
               d.Rect(rLeft, rTop, rRight, rBottom, shortStyle) + nl;
         }
      } else {
         // Logic site box
         boxes += d.Comment("Site[2]: " + siteType) + nl +
            d.Rect(rLeft, rTop, rRight, rBottom, style) + nl;
      }

      // Map over the INTs looking for globals.
      ScanForGlobals sfg(contourMap);
      deviceModel->MapIntTiles(sfg, site, tileDef);
      if (sfg.hasGlobal && 3 * globalInset < (rRight - rLeft)) {
         // Don't bother drawing if the site is very skinny, like the
         // 7 series CMT sites which divide a single column into four.
         // The global indicator will be covered by EmitTiles().
         boxes += d.Comment("Global inset") + nl +
            d.Rect(rLeft + globalInset,
                   rTop - globalInset,
                   rRight - globalInset,
                   rBottom + globalInset,
                   globalStyle) + nl;
      }

      if (gs) {
         boxes += d.Comment("Supersite Color") + nl +
            d.Rect(rLeft, rTop, rRight, rBottom,
                   d.Decl("fill", fillColor) +
                   d.Decl("fill-opacity", fillOpacity) +
                   d.Decl("stroke", "none"))
            + nl;
      }
      Rf_Cell superSiteCell(deviceModel->GetSiteParentTile(site));
      Tile* superTile(GetTile(superSiteCell));
      ASSERT(superTile);
      if ((!isConfigCenter && superTile->HasViolation()) ||
          (isConfigCenter && hasConfigCenterViolation))
      {
         // Crosshatch the supersite indicating a violation.
         boxes +=
            d.Comment(tile->Describe()) + nl +
            d.Rect(rLeft, rTop, rRight, rBottom, superSiteViolation)
            + nl;
      }

      if (isConfigCenter) {
         siteType = "CONFIGURATION";
      }
      AddLabel(rLeft,
               2 * offset + // Try to reduce collisions
               ((yTop + boxHeight - specialInset) +
                (yBottom + specialInset))/2.0,
               siteType, "");
   }
}

// The Spartan-6 Gigabit Transceivers are T-shaped. All the same
// indicators are required, but the SVG polygon drawing is needed
// rather than rectangles.
void DiagramEmitter::DrawNonRectangularSuperSite(Rf_Cell site,
                                                 string& boxes, string& labels)
{
   Svg d;
   if (IsDrawableSuperSite(site) && !IsRectangular(site)) {
      const Rf_CellDef tileDef(deviceModel->GetRfDevice()->GetTopCellDefTile());
      Rf_Cell cell(deviceModel->GetIntTile(site, tileDef, 1));
      Tile* tile(GetTile(cell));
      ASSERT(tile);
         
      // Not rectangular. Look up the boundary.
      const CoordVec& boundary(superBoundaryMap.find(site)->second);
      Data::GroupSet* gs(tile->GetGroupSet());
      string siteType(site.GetCellDef().GetName());
      boxes += d.Comment("Site[3]: " + siteType) + nl;
      if (gs && tile->HasProgramming()) {
         string color("none");
         if (!data.NeedConstraintsCheck()) {
            color = GetGroupColor(gs);
         }
         boxes += d.Polygon(boundary,
                            d.Decl("fill", color) +
                            d.Decl("fill-opacity", fillOpacity) +
                            d.Decl("stroke", "none"));
         boxes += d.Polygon(boundary, superSiteStyle);
      }
      if (tile->HasViolation()) {
         boxes += d.Comment(tile->Describe()) + nl +
            d.Polygon(boundary, superSiteViolation);
      }

      const BoundaryMap::const_iterator bmIt(superGlobalMap.find(site));
      if (bmIt != superGlobalMap.end()) {
         boxes += d.Polygon(bmIt->second, globalStyle);
      }

      // Align the label with the left edge of the perimeter,
      // centered between the top and bottom.
      ASSERT(1 < boundary.size());
      double xLeft(boundary.at(0));
      double xRight(boundary.at(0));
      double yTop(boundary.at(1));
      double yBottom(boundary.at(1));
      for (size_t k(2); k < boundary.size(); k += 2) {
         double x(boundary.at(k));
         double y(boundary.at(k + 1));
         if (x < xLeft) xLeft = x;
         if (xRight < x) xRight = x;
         if (y < yTop) yTop = y;
         if (yBottom < y) yBottom = y;
      }
      AddLabel((xLeft + xRight)/2.0,
               (yTop + yBottom)/2.0,
               siteType, "");
   }
}

// Draw the perimeter of every supersite, whether it is used or not,
// to help the user see the correspondence between PlanAhead, FPGA
// Editor, and IVT.
const string DiagramEmitter::EmitSuperSites() {
   Svg d;
   string box;
   string labels;
   Rf_CellContainer sites(deviceModel->GetSites());

   DrawClockColumn(box);

   // For each site
   for (Rf_CellContainer::const_iterator sitesIt(sites.begin());
        sitesIt != sites.end();
        ++sitesIt)
   {
      Rf_Cell site(*sitesIt);

      // Draw the IOB 'singles'. In 7-series IOBs come in two
      // sizes.
      DrawIOBSite(site, box, labels);

      // The previous condition and the following condition should be
      // mutually exclusive, but if they are ever not, better that both
      // sites are drawn.
      DrawRectangularSuperSite(site, box, labels);
      DrawNonRectangularSuperSite(site, box, labels);
   }

   labels = d.Elt("g", font, nl + EmitLabels() + nl);
   return d.Comment("Supersites") + nl +
      box + nl +
      labels + nl;
}

// Construct a string that describes a rounded rectangle outline with
// two strokes, one thick and one thin, preceded by a comment.
//
// Outline the outline so light colors don't disappear and so
// it is possible to distinguish the inside from the outside
// without any fill.
const string DiagramEmitter::AreaRangeRect(const string& comment,
                                           const string& groupColor,
                                           double xMin, double yMax,
                                           double xMax, double yMin)
{
   // Radius to center line is 
   //    radius1 + (stroke-width1 - stroke-width2)/2
   const double radius2(areaRangeCornerRadius +
                        (areaRangeCornerRadius - areaRangeOutlineWidth2)/2.0);
   Svg d;
   return 
      d.Comment(comment) + nl +
      d.Rect(xMin, yMax, xMax, yMin,
             d.Decl("stroke", groupColor),
             areaRangeCornerRadius)
      + nl +
      d.Rect(xMin - areaRangeOutlineWidth2,
             yMax + areaRangeOutlineWidth2,
             xMax + areaRangeOutlineWidth2,
             yMin - areaRangeOutlineWidth2,
             d.Decl("stroke", groupColor) +
             d.Decl("stroke-width", areaRangeOutlineWidth2) +
             d.Decl("stroke-opacity", 1.0) +
             d.Decl("pointer-events", "all"),
             radius2,
             comment) +
      nl;
}

// Draw a pair of rectangles with rounded corners, one slightly
// smaller than the other so that it is possible to distinguish inside
// from outside while looking at a magnified view.
void DiagramEmitter::EmitOneAreaRange(DeviceModel::SiteRect& rc,
                                      const string& group,
                                      const string& groupColor,
                                      bool isFirst,
                                      string& frames,
                                      string& labels)
{
   Svg d;
   // We found the interconnect tiles that correspond to
   // this site range, but we need to transform the
   // column account for hidden columns.

   // If the given corners are not the canonical ones,
   // we are still assured that lower is no higher than upper,
   // but we need to swap xMin and xMax.

   // As much as we'd like to use the canonical field, it applies to
   // the sites, not the interconnect associated with the sites, now
   // that interconnect can be arranged back-to-back.
   Rf_Cell leftInt(rc.lowerInt.GetTileColumn() <= rc.upperInt.GetTileColumn()
                   ? rc.lowerInt : rc.upperInt);
   Rf_Cell rightInt(leftInt == rc.lowerInt ? rc.upperInt : rc.lowerInt);

   ASSERT(leftInt.GetTileColumn() <= rightInt.GetTileColumn());
   double xMin(contourMap->AdjustX(leftInt.GetTileColumn())
               + GetSmallZynqOffset());
   // Site coordinates are tile-centric, but our
   // rectangle is edge-centric, so add 1 to include the
   // maximum tile in the boundary rectangle.
   double xMax(contourMap->AdjustX(rightInt.GetTileColumn()) + 1.0
               + GetSmallZynqOffset());

   ASSERT(xMin < xMax);

   // Swap rc.upperInt and rc.lowerInt because row numbers increase
   // downwards, but site numbers increase upwards.
   // Coordinates identify tiles, but to draw a rectangle
   // around a set of tiles, we need coordinates for the
   // edges of the set of tiles, hence yMax + 1.
   double yMin(contourMap->AdjustY(rc.upperInt.GetTileRow()));
   double yMax(contourMap->AdjustY(rc.lowerInt.GetTileRow()) + 1);

   // Compute the midpoint for the label location.
   double xMid((xMin + xMax)/2.0);
   double yMid((yMin + yMax)/2.0);

   if (isFirst) {
      string rect(AreaRangeRect("Group: " + group +
                                " Range: " + rc.rangeString,
                                groupColor,
                                xMin, yMax, xMax, yMin));
      if (rc.siteType == "SLICE") {
         frames = rect + frames;
      } else {
         frames += rect;
      }
   }

   // Label slice ranges with the isolation group. Label
   // other site ranges with the site type.
   string label;
   double textOpacity(0.75);
   double textSize(1.0);
   if (rc.siteType == "SLICE") {
      label = group;
   } else {
      label = rc.siteType;
      textOpacity = 0.25;
      textSize = 0.75;
   }
   string textStyle(d.Attr("fill-opacity", textOpacity)
                    + d.Attr("font-size", textSize));
   // Here's a hack to reduce label collisions. If this label is from
   // an area range that duplicates a previous area range, then offset
   // the label downwards by 1-5 lines using a ad hoc hashing
   // function. The offset is determined by the number of characters
   // in the site type string xor'd with the first two characters of
   // the label. This should help with RAMB8 vs RAMB16, which use the
   // same hardware, but in a different mode and help with the I/O
   // sites.
   double dupOffset((!isFirst) * lineHeightLarge *
                    ((rc.siteType.size() ^ label[0] ^ label[1]) % 5 + 1));
   labels += d.Text(xMid, yMid + dupOffset, label, textStyle) +
      nl;
}

// In order to suppress duplicate area range rectangles, a std::set is
// used to store the coordinates and color. Any consistent ordering
// will do.
typedef struct LexLess {
   bool operator()(const vector<size_t>& a, const vector<size_t>& b) const {
      return lexicographical_compare(a.begin(), a.end(), b.begin(), b.end());
   }
} LexLess;

// Add text and rectangles to the output indicating where the area
// range constraints are in relation to the tile array.  Strictly
// speaking, drawing the area ranges is not required, but it makes it
// easier to correlate the IVT report to FPGA Editor, PlanAhead, and
// the design source.
//
// Only area ranges for so-called user tiles are drawn. This
// substantially reduces the incidence of overlapping labels.
//
// Area ranges are drawn as rectangles with rounded corners. Both the
// outline color and the fill color are selected to indicate the
// isolation group. The label for each rectangle is drawn in the
// center of the rectangle. Slices are the only area range type that
// will always be present in an isolation group. Therefore slice
// ranges are labeled with the isolation group name. Other ranges are
// labeled with the site type.
const string DiagramEmitter::EmitAreaRanges() {
   Svg d;
   size_t colorIndex(ColorName_User);
   string frames;
   string labels;
   set<vector<size_t>, LexLess> dups;

   // For each site range...
   for (Data::SiteRectMapIt srmIt(data.siteRectMap.begin());
        srmIt != data.siteRectMap.end();
        ++srmIt)
   {
      Data::IsolationGroupName group(srmIt->first);
      Data::GroupSet* gs(groupSetFamily.FindOrMakeSet(group));
      string groupColor(GetGroupColor(gs));
      DeviceModel::SiteRectVec rangeList(srmIt->second);
      for (DeviceModel::SiteRectVecIt rlIt(rangeList.begin());
           rlIt != rangeList.end();
           ++rlIt)
      {
         DeviceModel::SiteRect rc(*rlIt);

         if (deviceModel->IsUserTileType(rc.siteType) &&
            rc.siteType != "PS7")
         {
            // We exclude the PS7 because all of its interconnect
            // blocks are on one side, therefore the bounding box
            // method does not work. We draw the area range along with
            // the supersite in DrawRectangularSuperSite() above.
            if (rc.lowerInt && rc.upperInt) {
               // Deduplicate the area range rectangles. Here we need
               // a canonical representation of the rectangle to
               // draw. For this purpose, it doesn't matter if the
               // hardware has sites at the corners or not, so we can
               // take x from one cell and y from another.
               size_t key[5] = {
                  rc.lowerInt.GetTileColumn(),
                  rc.upperInt.GetTileColumn(),
                  rc.lowerInt.GetTileRow(),
                  rc.upperInt.GetTileRow(),
                  colorIndex
               };
               if (key[0] < key[1]) {
                  key[0] = key[1];
                  key[1] = rc.lowerInt.GetTileColumn();
               }
               if (key[2] < key[3]) {
                  key[2] = key[3];
                  key[3] = rc.lowerInt.GetTileRow();
               }
               vector<size_t> vk(key, key + 5);
               bool isFirst(dups.insert(vk).second);
               EmitOneAreaRange(rc, group, groupColor, isFirst,
                                frames, labels);
            } else {
               // tileMin is rc.lowerInt and tileMax is rc.upperInt.
               Warn(100, "Unable to find both tileMin and tileMax.\n"
                    + rc.areaGroupName + ":" + rc.siteType + "\n"
                    "Contact Xilinx for assistance.");
            }
         }
      }
      if (colorIndex < ColorName_Overflow) {
         ++colorIndex;
      }
   }
   frames = d.Elt("g", d.Attr("style",
                              d.Decl("fill", "none") +
                              d.Decl("stroke-opacity", areaRangeStrokeOpacity) +
                              d.Decl("stroke-width", areaRangeStrokeWidth)),
                  nl + frames);
   labels = d.Elt("g", d.Attr("style",
                              d.Decl("text-anchor", "middle") +
                              d.Decl("dominant-baseline", "middle") +
                              d.Decl("font-family", "Helvetica")),
                  nl + labels);
   return d.Comment("Area Ranges") + nl +
      frames + nl +
      labels + nl;
}

// R#draw_fence - Generate an SVG diagram showing the a schematic
// floorplan of the design. Relative geometric relationships are
// preserved, but the drawing is not to scale and many details are
// hidden. The diagram shows the overall device bounds, site-based
// area ranges used to constrain the placement of the isolated logic,
// and interconnect tiles, used as the basis for isolation. Trusted
// Bus Macros are drawn with solid black bordered tiles. Tiles
// containing global signals are marked with an inset square . Only
// occupied tiles are drawn. Tiles containing isolated nets from a
// single isolation group are color coded.
void DiagramEmitter::GenerateDiagram() {
   LabelGroupSets();
   Svg d;
   const long cellSize(20);
   const long leftMargin(100); // Leave space for centered labels.
   const long topMargin(10);
   const long rightMargin(100); // Leave space for centered labels.
   const long bottomMargin(10);

   const long width(contourMap->AdjustX(contourMap->ColumnCount() - 1) + 1 +
                    GetSmallZynqOffset());
   const long height(contourMap->AdjustY(contourMap->RowCount() - 1) + 1);

   // Define an SVG pattern of dots as a simple grid of circles
   const string
      stipple(d.Elt("pattern",
                    d.Attr("id","stipple") +
                    d.Attr("x", 0.0) + d.Attr("y", 0.0) +
                    d.Attr("width", 0.1) + d.Attr("height", 0.1) +
                    d.Attr("patternUnits", "userSpaceOnUse") +
                    d.Attr("patternTransform", "rotate(26.5651)"),
                    nl +
                    d.Elt("circle",
                          d.Attr("cx", 0.0) +
                          d.Attr("cy", 0.0) +
                          d.Attr("r", 0.04) +
                          d.Attr("style", d.Decl("fill", "silver") +
                                 d.Decl("fill-opacity", 1.0) +
                                 d.Decl("stroke", "none"))) + nl));
   const string
      crosshatch(d.Elt("pattern",
                       d.Attr("id","crosshatch") +
                       d.Attr("x", 0.0) + d.Attr("y", 0.0) +
                       d.Attr("width", 0.5) + d.Attr("height", 0.5) +
                       d.Attr("patternUnits", "userSpaceOnUse") +
                       d.Attr("patternTransform", "rotate(45)"),
                       nl +
                       d.Line(-0.01, 0.25, 0.51, 0.25,
                              d.Decl("stroke", "black") +
                              d.Decl("stroke-width", 0.02) +
                              d.Decl("stroke-opacity", 1.0)) + nl + 
                       d.Line(0.25, -0.01, 0.25, 0.51,
                              d.Decl("stroke", "black") +
                              d.Decl("stroke-width", 0.02) +
                              d.Decl("stroke-opacity", 1.0)) + nl));
   const string
      horstripes(d.Elt("pattern",
                       d.Attr("id","horstripes") +
                       d.Attr("x", 0.0) + d.Attr("y", 0.0) +
                       d.Attr("width", 0.5) + d.Attr("height", 0.125) +
                       d.Attr("patternUnits", "userSpaceOnUse"),
                       nl +
                       d.Line(-0.01, 0.0625, 0.51, 0.0625,
                              d.Decl("stroke", "black") +
                              d.Decl("stroke-width", 0.01) +
                              d.Decl("stroke-opacity", 0.5)) + nl));
   const string
      vertstripes(d.Elt("pattern",
                       d.Attr("id","vertstripes") +
                       d.Attr("x", 0.0) + d.Attr("y", 0.0) +
                       d.Attr("width", 0.125) + d.Attr("height", 0.5) +
                       d.Attr("patternUnits", "userSpaceOnUse"),
                       nl +
                       d.Line(0.0625, -0.01, 0.0625, 0.51,
                              d.Decl("stroke", "black") +
                              d.Decl("stroke-width", 0.01) +
                              d.Decl("stroke-opacity", 0.5)) + nl));
   Draw(d.EmitSvg(nl
                  + stipple + nl
                  + crosshatch + nl
                  + horstripes + nl
                  + vertstripes + nl
                  + d.Elt("title", "", data.drawingTitle) + nl
                  + d.Elt("g",
                          d.Attr("transform",
                                 "translate("
                                 + ToString(leftMargin) + ","
                                 + ToString(topMargin)
                                 + ")scale("
                                 + ToString(cellSize) + ","
                                 + ToString(cellSize) + ")"),
                          d.Rect(0, height, width, 0) + nl
                          + EmitAreaRanges() + nl
                          + EmitTiles() + nl
                          + EmitSuperSites() + nl),
                  leftMargin + cellSize * width + rightMargin,
                  topMargin + cellSize * height + bottomMargin));
}
