// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: GroupSetFamily.c,v 1.5.206.2 2013/06/24 16:57:33 jcorbett Exp $
#include "GroupSetFamily.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.5.206.2 $";

// Constructor.
GroupSetFamily::GroupSetFamily() : ordered(false)
{
   // Typically there are only a few isolation groups. It is unlikely
   // that there would be more than a dozen
   groupSetList.reserve(64);
}

// See destructor below.
struct DeleteObject {
   template <typename T> void operator()(T* p) { delete p; }
};

// Delete each instance in groupSetList.
GroupSetFamily::~GroupSetFamily() {
   std::for_each(groupSetList.begin(), groupSetList.end(), DeleteObject());
}

// Return
//    Union({name}, nameSet) == extantSet
// In words, return true if an only if the set that would be formed by
// inserting name into nameSet equal extantSet.
//
// Note: nameSet may be null.
bool GroupSetFamily::Equal(const Data::IsolationGroupName& name,
                           const Data::GroupSet* nameSet,
                           const Data::GroupSet* extantSet) const
{
   ASSERT(extantSet);
   bool result(false);
   if (nameSet) {
      result =
         includes(extantSet->begin(), extantSet->end(),
                  nameSet->begin(), nameSet->end()) &&
         extantSet->find(name) != extantSet->end();
   } else {
      result = extantSet->size() == 1 && name == *extantSet->begin();
   }
   return result;
}

// Find the set that would be formed by inserting name in to nameSet,
// if it already exists.
//
// Linear search of an unordered list.
Data::GroupSet* GroupSetFamily::Find(const Data::IsolationGroupName& name,
                                     const Data::GroupSet* nameSet) const
{
   Data::GroupSet* result(0);
   for (GroupSetListConstIt i(groupSetList.begin());
        i != groupSetList.end() && !result;
        ++i)
   {
      if (Equal(name, nameSet, *i)) {
         result = *i;
      }
   }
   return result;
}

// Return a pointer to a set containing name and all the elements of
// nameSet, creating a new set, if necessary.
//
// If no such set exists in groupSetList, create a new set and add it to
// groupSetList.
//
// There are three cases:
//    1. If name is already in nameSet, return the address of nameSet.
//    2. If name + nameSet is already in groupSetList, return the address
//       of the union.
//    3. If name + nameSet is not in groupSetList,
//       create the union,
//       insert the union in groupSetList,
//       mark the list unsorted, and
//       return the address of the union.
Data::GroupSet* GroupSetFamily::FindOrMakeSet(const Data::IsolationGroupName&
                                                name,
                                              Data::GroupSet* nameSet)
{
   Data::GroupSet* result(nameSet);
   if (!nameSet || nameSet->find(name) == nameSet->end()) {
      // nameSet does not contain name. Do we already have a set that
      // equals {name} union nameSet?
      result = Find(name, nameSet);
      if (!result) {
         ASSERT(!ordered);
         // Need to create a set.
         result = new Data::GroupSet();
         if (nameSet) {
            copy(nameSet->begin(),
                 nameSet->end(),
                 insert_iterator<Data::GroupSet>(*result, result->begin()));
         }
         result->insert(name);
         groupSetList.push_back(result);
         ordered = false;
      }
   }
   return result;
}

// Predicate to find a set in a collection of sets.
class GroupSetMatch {
public:
   GroupSetMatch(const Data::GroupSet* target) : b(target) {}
   bool operator()(const Data::GroupSet* a) const
   {
      return *a == *b;
   }
private:
   const Data::GroupSet* b;
};

// Return the union of a and b, caching the result in groupSetList.
Data::GroupSet* GroupSetFamily::FindOrMakeUnion(Data::GroupSet* a,
                                                Data::GroupSet* b)
{
   Data::GroupSet* result(a);
   if (!a) {
      result = b;
   } else if (!b
              || a == b
              || includes(a->begin(), a->end(), b->begin(), b->end()))
   {
      // Set a is a superset of set b.
      ;
   } else if (includes(b->begin(), b->end(), a->begin(), a->end())) {
      // Set b includes set a, so result is b.
      result = b;
   } else {
      ASSERT(!ordered);
      // See if a union b is already in the group set list. If so,
      // return the pointer to the extant copy, otherwise add it and
      // return the new pointer.

      // Need a new set, at least temporarily.
      result = new Data::GroupSet();
      set_union(a->begin(), a->end(),
                b->begin(), b->end(),
                inserter(*result, result->begin()));
      GroupSetListConstIt begin(groupSetList.begin());
      GroupSetListConstIt end(groupSetList.end());
      GroupSetListConstIt it(find_if(begin, end, GroupSetMatch(result)));
      if (it == end) {
         // The union was not found. Add it.
         groupSetList.push_back(result);
         ordered = false;
      } else {
         // Already had the union. Delete the temporary and return the
         // cached one.
         delete result;
         result = *it;
      }
   }
   return result;
}

// Return true if and only if set a should be listed before set b in
// a ascending alphabetical order.
bool GroupSetLess(const Data::GroupSet* a, const Data::GroupSet* b) {
   return lexicographical_compare(a->begin(), a->end(), b->begin(), b->end());
}

// Sort groupSetList in ascending alphabetical order.
void GroupSetFamily::Sort() {
   sort(groupSetList.begin(), groupSetList.end(), GroupSetLess);
   ordered = true;
}
