// Copyright (C) 2009-2013 Xilinx, Inc.  All rights reserved
// $Id: PinChecker.c,v 1.44.206.4 2013/07/06 22:39:02 jcorbett Exp $

#include "Portability/export/Port_Type.h"
#include "Pds/export/Pdl/Pds_PdlExtendedPinType.h"
#include "Databases/PersonalityModule/export/baspmpic.h"
#include "Utilities/export/UtilSharedObject.h"

#include "main.h"
#include "PinGrid.h"
#include "PinChecker.h"
#include "ArgParser.h"

using namespace std;

// required for Port Utilities
static const char modname[] = __FILE__;
static const char rcsrev[] = "$Revision: 1.44.206.4 $";

PinChecker::PinChecker(Data& data) : data(data) {}

// Create a mapping from IO buffer cells to package pins.  A few
// typical package pin names: AF15, AF16, F14, and F13.
void PinChecker::LabelCells(CellToPackagePinMap& pkgMap)
{
   const Rf_Package* package(data.GetDeviceModel()->GetRfPackage());
   for (Rf_Package::Pins::const_iterator
           pinIt(package->begin());
        pinIt != package->end();
        ++pinIt)
   {
      const Rf_PackagePin& pkgPin(*pinIt);
      // In Virtex-4 and Virtex-5, the pad and the I/O buffer
      // are grouped together in 1 cell. In Spartan-6, the equivalent
      // circuitry is composed of a pad and several other tiles.
      const Rf_Cell pad(pinIt->GetCell());
      if (pkgMap.find(pad) != pkgMap.end()) {
         FatalInternal(103, "About to overwrite pkgMap entry.");
      }
      if (pad && pkgPin.IsExternal()) {
         pkgMap[pad] = pkgPin;
      } else if (pad) {
         // If a pad is not bonded, it would be reasonable to exclude
         // it from further adjacency analysis, but there isn't a
         // compelling need to perform the analysis that would justify
         // the exclusion. One might argue to the contrary noting that
         // in architectures such as Virtex-4 and Virtex-5, the
         // software device model combines I/O pads and I/O buffers;
         // it is possible to use the I/O buffer even though there is
         // no corresponding package pin. One might do this to gain
         // access to a register in the I/O buffer or to create a
         // design targeting multiple part packages.
         pkgMap[pad] = pkgPin;
      } else {
         // From the ccdoc for GetCell():
         //   "Many pins, eg power and ground, don't have cells."
         // I assume there is nothing to check for pins without cells.
      }
   }
}

// Find the group associated with sigPin and record it in pkgPinGrid
// and ioBufferGrid.
void PinChecker::LabelCompPin(CellToPackagePinMap& pkgMap,
                              XCad_CadBlock* pinBlock,
                              XCad_CadPin* topPin)
{
   // It's a pad, it's got a placement, so we can get its cell.
   // GetPlacement works on PADs, but not other pins on the top
   // level.
   Rf_Cell cell(pinBlock->GetPlacement());
   if (cell) {
      CellToPackagePinMap::iterator pkgIt(pkgMap.find(cell));
      if (pkgIt != pkgMap.end()) {
         const Rf_PackagePin& pkgPin(pkgIt->second);
         XCad_CadNetwork* net(topPin->GetNetwork());
         const string netName(net->GetReportableName());
         Data::NetGroupIt gIt(data.netGroup.find(net));
         string group("combined");
         if (gIt != data.netGroup.end()) {
            group = gIt->second;
         }
         pinGrid->LabelPin(pkgPin, netName, group);
      } else {
         const string cellName(cell.GetName());
         FatalInternal(107, "Unable to find " + cellName
                       + " in cell-to-package map.");
      }
   } else {
      FatalInput(4, "Encountered a pad on pin " + string(topPin->GetName())
                 + " with no cell.\n"
                 "Check that the design as placed and routed without\n"
                 "errors or warnings.");
   }
   ConsoleDot();
}

void PinChecker::LabelSpecialPin(CellToPackagePinMap& pkgMap,
                                 XCad_CadPin* topPin)
{
   // Here we are in the house that Jack built. We search from the
   // pin to the net to the signal to the other pin to the block
   // to the cell to the other other pin to the other net to the
   // group.
   XCad_CadNetwork* net(topPin->GetNetwork());
   if (!net) {
      const string pinName(topPin->GetName());
      FatalInternal(71, "Unable to get net for pin " + pinName + ".");
   }
   XCad_CadNetwork::Signals sigs(net->GetSignals());
   string group("[missing group]");
   bool found(false);
   for (XCad_CadNetwork::Signals::iterator sigIt(sigs.begin());
        sigIt != sigs.end() && !found;
        ++sigIt)
   {
      ConsoleDot();
      XCad_CadSignal::Pins pins((*sigIt)->GetPins());
      for (XCad_CadSignal::Pins::iterator subPinIt(pins.begin());
           subPinIt != pins.end() && !found;
           ++subPinIt)
      {
         XCad_CadPin* subPin(*subPinIt);
         // Note: subPin names are PAD, OUT, and PAD.

         Rf_Cell pad(subPin->GetBlock()->GetPlacement());
         if (pad) {
            // Note: In the current example, we find what we're looking
            // for only when the blockPin name is "I" or "O". I imagine
            // other examples could have other names like IO and maybe
            // some names related to differential IO standards or
            // tristated outputs, etc. So, ignore the name and try them all.
            XCad_CadBlock* pinBlock(subPin->GetBlock());
            XCad_CadBlock::Pins blockPins(pinBlock->GetPins());
            for (XCad_CadBlock::Pins::iterator blockPinIt(blockPins.begin());
                 blockPinIt != blockPins.end() && !found;
                 ++blockPinIt)
            {
               XCad_CadPin* blockPin(*blockPinIt);
               if (blockPin->GetNetwork()) {
                  string netName(blockPin->GetNetwork()->GetName());
                  for (Data::NetGroupIt gIt(data.netGroup.begin());
                       gIt != data.netGroup.end();
                       ++gIt)
                  {
                     if (netName == gIt->first->GetName()) {
                        CellToPackagePinMap::iterator
                           pkgIt(pkgMap.find(pad));
                        if (pkgIt != pkgMap.end()) {
                           const Rf_PackagePin& pkgPin(pkgIt->second);
                           const XCad_CadNetwork*
                              blockNet(blockPin->GetNetwork());
                           const string
                              netName(blockNet->GetReportableName());
                           group = gIt->second;
                           if (group == "ignored") {
                              // This happens for Zynq IOPADs. Since
                              // the nets connected to the pads are
                              // direct routing rather than general
                              // interconnect, their groups get
                              // changed to "ignored". Here we recover
                              // the group based on the block hierarchy.
                              XCad_CadBlock* blockParent(pinBlock->GetParent());
                              if (blockParent) {
                                 const string blockGroup(data.GetIsolationGroup(blockParent));
                                 if (blockGroup != "") {
                                    group = blockGroup;
                                 }
                              }
                           }
                           found = true;
                           pinGrid->LabelPin(pkgPin, netName, group);
                        }
                     }
                  }
               }
            }
         }
      }
   }
   if (!found) {
      const string& pinName(ToString(topPin->GetName()));
      Warn(75, "Unable to find isolation group for pin " + pinName
           + ".\nCheck the ngdbuild report for related errors or warnings.");
   }
}

void PinChecker::FindPinGroup(CellToPackagePinMap& pkgMap,
                              XCad_CadPin* topPin,
                              XCad_CadPin* sigPin)
{
   // We want to get from the pin on the top level block to the
   // cell that corresponds to a package pin.

   if (pkgMap.empty()) {
      FatalInternal(42, "pkgMap was not initialized.");
   }

   XCad_CadBlock* pinBlock(sigPin->GetBlock());
   XCad_CadBlock::BLOCKFUNCTIONALTYPE ft(pinBlock->GetFunctionalType());
   XCad_CadBlock::HIER_LEVEL hl(pinBlock->GetHierLevel());

   if (ft == XCad_CadBlock::FUNCTION_IO && hl == XCad_CadBlock::LEVEL_COMP) {
      LabelCompPin(pkgMap, pinBlock, topPin);
   } else if (ft == XCad_CadBlock::FUNCTION_SPECIAL ||
              (ft == XCad_CadBlock::FUNCTION_LOGIC &&
               hl == XCad_CadBlock::LEVEL_COMP))
   {
      LabelSpecialPin(pkgMap, topPin);
   } else {
      // In the unlikely event the BLOCKFUNCTIONALTYPE enum or
      // HIER_LEVEL enum are changed, the diagnostic information won't
      // be accurate, but it's easy enough to figure out the
      // integer value of the enumerations and cross-reference these
      // with the current values.
      const char* blockFunctionalTypeName[] = {
         "FUNCTION_LOGIC", "FUNCTION_IO", "FUNCTION_SPECIAL"
      };
      const char* hierLevelName[] = {
         "LEVEL_ROOT", "LEVEL_MODULE", "LEVEL_MACRO", "LEVEL_COMP",
         "LEVEL_BEL", "LEVEL_UNKNOWN"
      };
      string type(blockFunctionalTypeName[ft]);
      string level(hierLevelName[hl]);
      const string pinName(sigPin->GetName());
      FatalInternal(26, "Pin " + pinName +
                    " has unexpected functional type or hierarchical level.\n"
                    "  Functional Block Type = " + type + "\n"
                    "  Hierarchical Level = " + level + "\n"
                    "Group cannot be determined.");
   }
}


// For each pin on the top level design block, find the cell of its
// IOB if any. Use that cell to look up the package pin.
void PinChecker::LabelPins(CellToPackagePinMap& pkgMap)
{
   if (ArgParser::verbose) {

      Section("Package Pins, I/O Buffers, and I/O Banks");

      Report("  Note: It is incumbent on the user to verify that pins "
             "connected\n"
             "  to ignored nets are correct. For example, pins may not be\n"
             "  directly connected to trusted bus macros, but pins may be\n"
             "  connected to clocks, power, and global resets.\n\n"
             "    Pin(col, row) Bank      I/O Buffer Isolation Group Net\n"
             "  --------------- ---- --------------- --------------- -------");
   }

   XCad_CadView* view(data.groupData.find("combined")->second->GetView());
   XCad_CadBlock* block(view->GetBlock());

   typedef XCad_CadBlock::Pins PinVec;
   PinVec pins(block->GetPins());
   for (PinVec::iterator pinIt(pins.begin()); pinIt != pins.end(); ++pinIt) {
      XCad_CadPin* topPin(*pinIt);

      // Note: These all claim to be of type pad.
      // topPin->GetExtendedPinType().GetPinType()

      XCad_CadSignal* signal(topPin->GetChildSignal());
      if (signal) {
         XCad_CadSignal::Pins pins(signal->GetPins());
         for (XCad_CadSignal::Pins::iterator sigPinIt(pins.begin());
              sigPinIt != pins.end();
              ++sigPinIt)
         {
            ConsoleDot();
            XCad_CadPin* sigPin(*sigPinIt);
            if (sigPin != topPin) {      // other end
               // Grab any pin on the block except the pin we came
               // from. We need the block of the child pin and all
               // pins are on the same block.
               FindPinGroup(pkgMap, topPin, sigPin);
               break;
            }
         }
      } else {
         // A pin with nothing connected to it. Not a problem for
         // isolation, but the customer may have something to fix.
         const string& pinName(ToString(topPin->GetName()));
         Warn(49, "Pin " + pinName + " appears unconnected.");
      }
   }
}

// Adjacent pins must be in the same group. This turns out to be a bit
// convoluted to verify because we have to trace through many objects
// to connect net names to package pins and device pins and we
// also have several different coordinate systems for cells, slices
// and I/O buffers (IOBs). It's a little like The House that Jack Built.
//    This is the farmer sowing the corn,
//    That kept the cock that crowed in the morn,
//    That waked the priest all shaven and shorn,
//    That married the man all tattered and torn,
//    That kissed the maiden all forlorn,
//    That milked the cow with the crumpled horn,
//    That tossed the dog,
//    That worried the cat,
//    That killed the rat,
//    That ate the malt
//    That lay in the house that Jack built.
void PinChecker::Check() {
   pinGrid.reset(new PinGrid(data.GetDeviceModel()));

   CellToPackagePinMap pkgMap;
   LabelCells(pkgMap);
   LabelPins(pkgMap);
   pinGrid->CheckPinLocations(); // R#check_package_pins,
                                 // R#report_io_assignments,
                                 // R#check_interconnect_routing,
                                 // R#check_direct_routing,
                                 // R#check_logic_tiles
   pinGrid->CheckIoBanks();      // R#check_io_banks
}

// Note: pinGrid will evaluate to 0 if no check has been performed.

// If the grid has been initialized, return its status.
int PinChecker::GetStatus() {
   if (pinGrid.get() != NULL) {
      return pinGrid->GetStatus();
   } else {
      return 0;
   }
}

// Write the pin faults to the report file. Note: It is an error to
// ask for a pin fault report if the Check() function that initializes
// the grid has not been called.
//
// R#check_package_pins, R#check_io_banks
// R#check_interconnect_routing, R#check_direct_routing,
// R#check_logic_tiles, R#check_io_banks
void PinChecker::ReportFaults() {
   ASSERT(pinGrid.get() != NULL);
   pinGrid->ReportFaults();
}

// Get the error count from the pin grid, otherwise no errors yet.
int PinChecker::GetErrorCount() {
   if (pinGrid.get() != NULL) {
      return pinGrid->GetErrorCount();
   } else {
      return 0;
   }
}
